"""Automated embedding of resources in Python Packages/Modules

ResourcePackage provides the Python programmer with a
mechanism for automatically maintaining a collection of
resources embedded in Python source-code modules.

The basic approach taken is to embed one resource in
each python module of a package explicitly intended
to be a resource-only package.

See the doc sub-directory for documentation
"""
__version__ = (1,0,0)
__author__ = "Michael Colin Fletcher"
__license__ = "BSD-Style, see license.txt for details"




import os

for testpackage in ('..\\testres','..\\testwx'):
	for file in os.listdir( testpackage ):
		if (
			os.path.splitext( file )[1].lower() == '.py' and
			file != '__init__.py'
		):
			f = os.path.join( testpackage, file )
			print 'del', f
			os.remove( f )



"""Objects for doing generic data-as-code encoding"""
import string, cStringIO, zlib, types, os
import resourcepackage

HEADER = '''# -*- coding: ISO-8859-1 -*-
"""Resource %(module)s (from file %(source)s)"""
# written by resourcepackage: %(resourcepackagev)r
source = %(source)r
package = %(packagen)r
'''
class SimpleGenerator:
	"""Simplistic generator"""
	def __repr__( self ):
		return "%s (singleton)"%( self.__class__.__name__)
	def __call__( self, source, destination, package=None ):
		"""Encode source in destination for package"""
		header = self.getHeader( source, destination, package )
		assert isinstance( header, types.StringType ), """Generator %s didn't return a String, returned %r"""%( self, header )
		data = self.getDataRepr( source, destination, package )
		assert isinstance( data, types.StringType ), """Generator %s didn't return a String, returned %r"""%( self, data )
		footer = self.getFooter( source, destination, package )
		assert isinstance( footer, types.StringType ), """Generator %s didn't return a String, returned %r"""%( self, footer )
		
		file = open(destination,'w')
		try:
			file.write( header )
			file.write( data )
			file.write( footer )
		finally:
			file.close()

	def getHeader( self, source, destination, package=None ):
		"""Get the header, written before the data variable"""
		source = os.path.basename( source )
		module = os.path.splitext(os.path.basename( destination ))[0]
		resourcepackagev = resourcepackage.__version__
		packagen = package.packageName
		return HEADER % locals()
		
	def getDataRepr( self, source, destination, package=None ):
		"""data as Python code assigning a value to variable "data" """
		return """data = %s\n"""%(
			crunch_data(
				self.getData(source)
			)
		)

	def getFooter( self, source, destination, package=None ):
		"""Get the footer, written after the data variable"""
		return """### end\n"""
	
	def getData( self, source, package=None ):
		"""Get the data to be encoded in the package"""
		return open(source,'rb').read()
		

class CompressedGenerator( SimpleGenerator ):
	"""Adds Zlib compression and decompression to generator"""
	def getDataRepr( self, source, destination, package=None ):
		"""data as Python code assigning a value to variable "data" """
		return """data = zlib.decompress(%s)\n"""%(
			crunch_data(
				self.getData(source)
			)
		)
	def getData( self, source, package=None ):
		"""Get the data to be encoded in the package"""
		data = SimpleGenerator.getData( self, source, package )
		return zlib.compress( data, 9 )
	def getHeader( self, source, destination, package=None ):
		"""Get the header, written before the data variable"""
		base = SimpleGenerator.getHeader( self, source, destination, package )
		return base + """\nimport zlib\n"""


_char_map = {
}
for c in range(32):
	_char_map[ chr(c)] = '\\%03o'%(c)
_char_map['"'] = '\\"'
_char_map['\\'] = '\\\\'

def crunch_data(source, chunkSize=60, charMap = _char_map ):
	"""Try to get a minimal representation of a binary as Python code

	This is a re-implementation that tries to
	minimise the size of the data by embedding extended
	characters in the source code.
	"""
	result = []
	lastIndex = 0
	for index in xrange(0, len(source), chunkSize):
		sresult = "".join([
			charMap.get( char, char)
			for char in source[index:index+chunkSize]
		])
		result.append( sresult )
	return '"%s"'%( '\\\n'.join( result ))

SIMPLE = SimpleGenerator()
COMPRESSED = CompressedGenerator()

generators = {
	"": SIMPLE,
	".txt": COMPRESSED,
	".html": COMPRESSED,
	".htm": COMPRESSED,
	".pdf":COMPRESSED,
}




#!/usr/bin/env python
"""Script for extracting resources from a resource package"""

usage = """extract.py [-f] packageName [modules, ...]

packageName -- dotted python package name for the package
	which holds the resource to be extracted.  The package
	__init__.py file should include a Package instance named
	"package".  If not, a new default package instance will
	be created for the given package.
	
modules -- optional list of individual modules to be
	extracted, otherwise scans the package directory
	to determine the modules to extract.

-f -- flag to signal that extraction should be "forced",
	that is, that the relative dates of the module and any
	existing resource file should be ignored, and the
	extraction should always occur.

Note:
	Because the extraction process needs to import the
	package, any automatic scanning done by your __init__.py
	may cause modules to be updated from their resources.
	
	As a result, if there is a newer resource file which
	would generate the module you are seeking to extract,
	the module will be overwritten before the extraction
	and you'll get a copy of the already-present resource
	file.  To avoid this, delete all copies of the resource
	files you would like to extract before running extract.py

"""
import os, string

def main( packageName, modules=(), force=0):
	"""Perform the actual extraction"""
	packageModule = __import__(
		packageName, {}, {},
		string.split(packageName, '.')
	)
	if not hasattr( packageModule, 'package' ):
		# build the default package object...
		from resourcepackage import package
		packageObject = package.Package(
			packageModule.__name__,
			directory = os.path.dirname( os.path.abspath(packageModule.__file__) ),
		)
	else:
		packageObject = packageModule.package
	if modules:
		for module in modules:
			packageObject.extractFile( module, force )
	else:
		packageObject.extract( force )

if __name__ == "__main__":
	import sys
	arguments = sys.argv[1:]
	try:
		import logging
		# this is wrong, should be done at the app level :( 
		log = logging.getLogger( "resourcepackage" )
		log.setLevel( logging.INFO )
		logging.basicConfig()
	except ImportError:
		log = None
	if arguments:
		# do we have a force...
		force = arguments[0] == '-f'
		if force:
			del arguments[0]
		if arguments:
			packageName = arguments[0]
			modules = arguments[1:]
			main( packageName, modules, force=force )
		else:
			print usage
			print 'ERR: arguments received', sys.argv[1:]
	else:
		print usage
			
	

		


# -*- coding: ISO-8859-1 -*-
"""Resource _cvsignore (from file .cvsignore)"""
# written by resourcepackage: (1, 0, 0)
source = '.cvsignore'
package = 'resourcepackage.testwx'
data = "*.pyc\015\012*.pyo\015\012*.max\015\012*.db\015\012*.bat\015\012*.cvsignore\015\012*.cache\015\012MANI\
FEST\015\012\015\012"
### end



"""wxPython-specific data-as-code extensions to defaultgenerators"""
from resourcepackage import defaultgenerators
from wxPython.wx import *
from wxPython.tools import img2img
import tempfile, os

formats = [
	'.png', '.jpg', '.ico', '.tif', '.bmp',
	'.xpm', '.gif', '.pcx', '.pnm', '.iff',
	'.jpeg', '.tiff',
]

HEADER = '''
### wxPython specific functions
originalExtension = %(extension)r
from wxPython.wx import wxImageFromStream, wxBitmapFromImage, wxEmptyIcon
import cStringIO
def getData( ):
	"""Return the data from the resource as a simple string"""
	return data
def getImage( ):
	"""Return the data from the resource as a wxImage"""
	stream = cStringIO.StringIO(data)
	return wxImageFromStream(stream)
def getBitmap( ):
	"""Return the data from the resource as a wxBitmap"""
	return wxBitmapFromImage(getImage())
def getIcon( ):
	"""Return the data from the resource as a wxIcon"""
	icon = wxEmptyIcon()
	icon.CopyFromBitmap(getBitmap())
	return icon
'''
class wxImageGenerator( defaultgenerators.SimpleGenerator ):
	"""wxPython specific handling of Image formats"""
	def getHeader( self, source, destination, package=None ):
		"""Get the header, written before the data variable"""
		baseName, extension = os.path.splitext( source )
		source = baseName + '.png'
		base = defaultgenerators.SimpleGenerator.getHeader(
			self, source, destination, package
		)
		return base + (HEADER%locals())
		
	def getData( self, source, package=None ):
		"""Get the image as a .png-encoded string

		This is adapted directly from img2py.py
		"""
		tfname = tempfile.mktemp()
		ok, msg = img2img.convert(
			source,
			None, # don't support resetting the mask value
			None,
			tfname,
			wxBITMAP_TYPE_PNG,
			".png",
		)
		if not ok:
			raise ValueError ("""Unable to convert file %s to a png image for %s: %s"""%( source, package, msg))
		fh = open(tfname, "rb")
		try:
			data = fh.read()
		finally:
			fh.close()
			os.remove( tfname )
		return data
	

GENERATOR = wxImageGenerator()

generators = {
}
for format in formats:
	generators[ format ] = GENERATOR



#!/usr/bin/env python
"""Installs ResourcePackage using distutils

Run:
	python setup.py install
to install the package from the source archive.
"""

if __name__ == "__main__":
	import sys,os, string
	from distutils.sysconfig import *
	from distutils.core import setup,Extension
	from distutils.command.build_ext import build_ext
	from distutils.command.install import install
	from distutils.command.install_data import install_data
	##from my_install_data import *

	##############
	## Following is from Pete Shinners,
	## apparently it will work around the reported bug on
	## some unix machines where the data files are copied
	## to weird locations if the user's configuration options
	## were entered during the wrong phase of the moon :) .
	from distutils.command.install_data import install_data
	class smart_install_data(install_data):
		def run(self):
			#need to change self.install_dir to the library dir
			install_cmd = self.get_finalized_command('install')
			self.install_dir = getattr(install_cmd, 'install_lib')
			# should create the directory if it doesn't exist!!!
			return install_data.run(self)
	##############
	### The following automates the inclusion of files while avoiding problems with UNIX
	### where case sensitivity matters.
	dataFiles = []
	excludedTypes = ('.py','.pyc','.pyo', '.db', '.max','.gz','.bat')
	def nonPythonFile( file ):
		if string.lower( file ) == 'cvs':
			return 0
		else:
			return (os.path.splitext( file )[1]).lower() not in excludedTypes
	dataDirectories = [
		'.',
		'testres',
	]
	for directory in dataDirectories:
		finalFiles = []
		for file in os.listdir( directory):
			fullFile = os.path.join( directory, file )
			if os.path.isfile(fullFile) and nonPythonFile(fullFile):
				finalFiles.append (os.path.join(directory, file))
		if finalFiles:
			dataFiles.append (
				(os.path.join('resourcepackage',directory),finalFiles)
			)
	### Now the actual set up call
	setup (
		name = "ResourcePackage",
		version = "1.0.0",
		description = "Automated resource-as-package embedding mechanism",
		author = "Mike C. Fletcher",
		author_email = "mcfletch@users.sourceforge.net",
		url = "http://resourcepackage.sourceforge.net",
		license = "BSD-style, see license.txt for details",

		package_dir = {
			'resourcepackage':'.',
		},

		packages = [
			'resourcepackage', 
			'resourcepackage.testres',
			'resourcepackage.testwx',
		],
		# non python files of examples      
		data_files = dataFiles,
		cmdclass = {'install_data':smart_install_data},
	)
	



"""Script to automatically generate OpenGLContext documentation"""
from OpenGLContext.pydoc import pydoc2

if __name__ == "__main__":
	excludes = [
		"wxPython.wx",
		"Numeric",
		"_tkinter",
		"Tkinter",
		"math",
	]
	stops = [
##		"resourcepackage.scanning__init__",
	]

	modules = [
		"resourcepackage",
	]	
	pydoc2.PackageDocumentationGenerator(
		baseModules = modules,
		destinationDirectory = "../doc/pydoc",
		exclusions = excludes,
		recursionStops = stops,
	).process ()



"""Package object, manages package-related operations
"""
import os, string, stat, traceback, types
from resourcepackage import defaultgenerators
try:
	import logging
	log = logging.getLogger( "resourcepackage" )
except ImportError:
	log = None

class Package:
	"""Represents an individual ResourcePackage at design-time

	Includes facilities for scanning (embedding) and extracting
	the package.
	"""
	directory = ""
	def __init__(
		self,
		packageName,
		directory,
		generators = defaultgenerators.generators,
	):
		"""Initialse the Repository

		directory -- the package's directory, in which we will
			do all of our work.
		"""
		self.packageName = packageName
		self.directory = directory
		self.generators = generators
	def __repr__( self ):
		return """%s (%s)"""%( self.__class__.__name__, self.packageName)

	# all-lowercased extensions and files to ignore
	ignoreExtensions = [ '.pyc','.pyo',]
	ignoreFiles = [ '__init__.py', ]

	def fileToName( self, filename ):
		"""Get Python module name and extension from filename

		This is a primary customisation point, as it allows you
		to choose the module-names in the package as generated
		by the system.  The default implementation does the
		following to the name:

			replaces all '.' with '_'
			replaces all ' ' with '_'

		So, obviously, there are cases where two different
		filenames will produce the same module name.  Those
		cases will be caught by the processing code and raise
		ValueErrors.

		Note that for case-in-sensitive file-systems, there
		will potentially be silent overwrites as files which
		differ only in case will not raise a ValueError.

		returns (base, extension)

			base -- processed filename
			extension -- lowercase-only version of the file
				extension, or "" if no extension was found.
		"""
		try:
			base, ext = os.path.splitext(filename)
		except ValueError:
			# there is no extension on the file...
			base, ext = filename, ''
		ext = string.lower(ext)
		# Alpha 3 changes, makes the module names less
		# consistent, but allows for more file-names to map
		# to unique module names
		#   Upper/lower-case is a tricky one,
		#   although I think they should be unified
		#   I'm guessing other's will want them unique
		#   base = string.lower( base )
		base = string.replace( filename, ".", "_" )
		base = string.replace( base, " ", "_" )
		return base, ext
		

	def isResource( self, file, ext ):
		"""Determine whether we consider this file a resource"""
		if not os.path.isfile( os.path.join( self.directory, file )):
			return 0
		file = string.lower( file )
		if (
			( ext not in self.ignoreExtensions ) and
			( file not in self.ignoreFiles ) and
			( ext != '.py' )
		):
			return 1
		return 0
			
	def scan( self, force=0 ):
		"""Scan the directory, looking for updated/added resources"""
		if log:
			log.info("""scan(force=%r) %s""", force, self )
		fileList = os.listdir( self.directory )
		nonPython = {}
		python = {}
		testFileNames = {}
		if log:
			log.debug("""filtering filename-list""" )
		for file in fileList:
			base, ext = self.fileToName( file )
			fullName = os.path.join( self.directory, file )
			if log:
				log.debug("""file=%r, base=%r, ext=%r""", file, base, ext )
			if self.isResource( file, ext ):
				if testFileNames.has_key( base ):
					raise ValueError(
						"""%s has two data files %s and %s which would generate the same Python module %s"""%(
							self, file, testFileNames[base]
						)
					)
				testFileNames[base] = file
				if log:
					log.debug("""will process""" )
				nonPython[file] = base, ext, fullName
			elif ext == '.py':
				python[base] = file, ext, fullName
		if log:
			log.debug("""starting updates, %s files""", len(nonPython) )
		for file,(base,ext,fullName) in nonPython.items():
			self.scanFile(
				file,
				base,
				extension=ext,
				force=force
			)
		if log:
			log.debug("""finished updates""")
	def scanFile( self, source, base=None, extension=None, force=0):
		"""Encode/update/scan a single file

		source -- full path name of the source-file
		base -- (optional) calculated base module name
		extension -- (optional) calculated lower-cased extension
		force -- whether to force update even if dates suggest
			the module is already up-to-date.

		Note: this does _not_ check to see if there is a filename
		conflict between resources, so potentially it could
		overwrite another source-file's destination module.
		"""
		if log:
			log.debug("""scanFile %r""", source )
		# since this can be called from scripts, not just
		# the scan method, we do a lot of redundant name-
		# mangling :(

		# This forced truncation makes some things possible,
		# and others not, for instance, you can't update into a
		# different directory from where the files are, but you
		# can specify the filenames as p:\whatever\whenever.gif
		# which in the imagined approach will be more common.
		source = os.path.basename( source )
		fullName = os.path.join( self.directory, source )
		if base is None or extension is None:
			base, extension = self.fileToName( source )
		moduleName = base + '.py'
		fullModuleName = os.path.join(self.directory, moduleName)
		
		# date-only check, is actually a redundant check for
		# calls from scan
		if (not force) and os.path.exists( fullModuleName ):
			if self.compareDates( fullName, fullModuleName ):
				if log:
					log.info("""file %r up to date""", moduleName )
				return 0
			else:
				if log:
					log.info("""refresh %r -> %r""", source, moduleName)
		elif force:
			if log:
				log.info("""force %r -> %r""", source, moduleName)
		else:
			if log:
				log.info("""new %r -> %r""", source, moduleName)
		# okay, one way or another we want to generate our
		# little Python file for this resource.  By default,
		# we're going to just dump the contents to a string.
		# a child can handle image-specific mechanisms...
		generator = self.getGenerator( extension )
		if log:
			log.debug("""generator %r""", generator )
		generator( fullName, fullModuleName, self )
		if log:
			log.debug("""scanFile %r finished""", source )
		return 1

	def getGenerator( self, extension="" ):
		"""Get a file-type-specific generator, or the default"""
		for ext in [ extension, "" ]:
			obj = self.generators.get( ext )
			if obj:
				return obj
		raise KeyError( """%s class has no default "generate" object (key "")!!!"""%(self.__class__))
	
				
	def compareDates( self, first, second ):
		"""Compare dates for two filenames

		This is it's own function so that a better
		comparison mechanism could be included easily.

		Returns boolean saying whether the first date is
		before-or-equal to the second
		"""
		try:
			firstTime = os.stat(first)[stat.ST_MTIME]
		except OSError:
			# first doesn't exist, so it's staler than
			# second for us...
			return 1
		try:
			secondTime = os.stat(second)[stat.ST_MTIME]
		except OSError:
			# second doesn't exist, so it's staler than
			# first for us...
			return 0
		return firstTime <= secondTime
	### Extracting back to disk files
	def isEncodedResource (self, file, ext ):
		"""Determine whether we consider this file an encoded resource module
		"""
		if not os.path.isfile( os.path.join( self.directory, file )):
			return 0
		if ext != '.py' or file == "__init__.py":
			return 0
		## OK, so it is definitely a python file...
		## for now, we will consider that sufficient...
		return 1
	def extract( self, force=0 ):
		"""Extract encoded files to disk files

		for each python source file other than __init__.py, we
		attempt to write:
			module.data to module.source

		if module.source doesn't exist, or is older,
		or force is true, then we write the file, otherwise
		we skip the file.
		"""
		if log:
			log.info("""scan(force=%r) %s""", force, self )
		fileList = filter ( self.isEncodedResource, os.listdir( self.directory ))
		for file in fileList:
			self.extractFile (module, force)

	def extractFile( self, module, force = 0 ):
		"""Extract a single file from source (python module) to destination"""
		# first question, should we do anything
		fullModule = os.path.join( self.directory, module)
		### Need to import the module to do anything...
		base, ext = self.fileToName( module )
		try:
			moduleObject = self.importModule(base)
		except Exception, err:
			if log:
				log.error( """Exception while attempting to extract module %s, %s""", module, err)
			return
		if hasattr( moduleObject, "data") and hasattr( moduleObject, "source"):
			fullDestination = os.path.join( self.directory, moduleObject.source)
			if force or not self.compareDates( fullModule, fullDestination):
				# okay, should write to the resource file...
				if log:
					log.info( """extract %s -> %s""", module, moduleObject.source)
				if isinstance( moduleObject.data, types.StringType ):
					fileHandle = open( fullDestination, 'wb')
					try:
						fileHandle.write( moduleObject.data )
					finally:
						fileHandle.close ()
					return 1
				else:
					if log:
						log.error( """module %s data attribute is not a string, is a %s""", module,type(moduleObject.data))
			else:
				if log:
					log.info( """resource file %s up-to-date""", fullDestination)
				
		else:
			if log:
				log.error( """Resource module %s does not define "data" and "source" attributes, cannot extract""", module)
			
	def importModule(self, baseName):
		"""Import the given module from our package and return the module object"""
		moduleName = string.split( self.packageName,'.')+[baseName]
		return __import__( string.join( moduleName, '.'), {}, {}, moduleName)

##
##if log:
##	log.setLevel( logging.WARN )
	


#!/usr/bin/env python
"""Script for scanning/updating resources into a resource package"""

usage = """scan.py [-f] packageName [filenames, ...]

packageName -- dotted Python package name for the package
	to be scanned.  If the Python package __init__.py
	includes a package instance named "package" it will
	be used for extraction.  If not, a new default
	package instance will be created for the given package.
	
filenames -- optional list of individual files to be
	updated, otherwise scans the package directory
	to determine the modules to create/update.

-f -- flag to signal that updates should be "forced",
	that is, that the relative dates of the resource file
	and any existing module should be ignored, and the
	embedding/encoding/updating should always occur.

Note:
	Because the scanning process needs to import the
	package, any automatic scanning done by your __init__.py
	may cause modules to be updated from their resources
	before scan.py starts working.

"""
import os, string

def main( packageName, filenames=(), force=0):
	"""Perform the actual extraction"""
	packageModule = __import__(
		packageName, {}, {},
		string.split(packageName, '.')
	)
	if not hasattr( packageModule, 'package' ):
		# build the default package object...
		from resourcepackage import package
		packageObject = package.Package(
			packageModule.__name__,
			directory = os.path.dirname( os.path.abspath(packageModule.__file__) ),
		)
	else:
		packageObject = packageModule.package
	if filenames:
		for filename in filenames:
			packageObject.scanFile( source = filename, force=force )
	else:
		packageObject.scan( force=force )

if __name__ == "__main__":
	import sys
	arguments = sys.argv[1:]
	try:
		import logging
		# this is wrong, should be done at the app level :( 
		log = logging.getLogger( "resourcepackage" )
		log.setLevel( logging.INFO )
		logging.basicConfig()
	except ImportError:
		log = None
	if arguments:
		# do we have a force...
		force = arguments[0] == '-f'
		if force:
			del arguments[0]
		if arguments:
			packageName = arguments[0]
			modules = arguments[1:]
			main( packageName, modules, force=force )
		else:
			print usage
			print 'ERR: arguments received', sys.argv[1:]
	else:
		print usage
			
